/*
 *                            *
 *                           * *
 *                          *   *
 *                     ***************
 *                      * *       * *
 *                       *  MUMPS  *
 *                      * *       * *
 *                     ***************
 *                          *   *
 *                           * *
 *                            *
 *
 *   fma_globals.c
 *    fmadm - globals
 *
 *  
 *   Author: John P. Willis <jpw@coherent-logic.com>
 *    Copyright (C) 1998 MUG Deutschland
 *    Copyright (C) 2020 Coherent Logic Development LLC
 *
 *
 *   This file is part of FreeM.
 *
 *   FreeM is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU Affero Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   FreeM is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU Affero Public License for more details.
 *
 *   You should have received a copy of the GNU Affero Public License
 *   along with FreeM.  If not, see <https://www.gnu.org/licenses/>.
 *
 **/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "fmadm.h"

void gl (char *global, short kf, short df, short nr, int chop);

int fma_globals_list (int optc, char **opts)
{
    DIR *dir;
    struct dirent *ent;

    char filename[STRLEN];

    int ct = 0;

    printf ("\nFreeM Global Listing\n");
    printf ("--------------------\n\n");

    printf ("Namespace:     %s\n", fma_namespace);
    printf ("Global Path:   %s\n\n", fma_global_path);

    if ((dir = opendir (fma_global_path)) == NULL) {
        fprintf (stderr, "fmadm:  could not open global directory %s\n", fma_global_path);
        return 1;
    }

    while ((ent = readdir (dir)) != NULL) {

        strncpy (filename, ent->d_name, STRLEN - 1);

        if (filename[0] == '^' && filename[1] != '$') {
            printf (" %s\n", filename);        
            ct++;
        }
        
    }

    printf ("\n\n    - %d globals found\n\n", ct);
    closedir (dir);

    return 0;
}

int fma_globals_examine (int optc, char **opts)
{

    DIR *dir;
    struct dirent *ent;
    
    char gpath[STRLEN];
    int i;
    int ct = 0;

    if ((dir = opendir (fma_global_path)) == NULL) {
        fprintf (stderr, "fmadm:  could not open global directory %s\n", fma_global_path);
        return 1;
    }

    
    printf ("\nFreeM Global Examine\n");
    printf ("--------------------\n\n");

    printf ("Namespace:         %s\n", fma_namespace);
    printf ("Global Path:       %s\n", fma_global_path);
    printf ("Globals Selected:  ");

    if (optc > 1) {
    
        for (i = 1; i < optc; i++) {
            printf ("%s ", opts[i]);
        }

        printf ("\n\n");
    
    }
    else {
        printf ("[ENTIRE NAMESPACE]\n\n");
    }


    if (optc > 1) {

        for (i = 1; i < optc; i++) {

            snprintf (gpath, STRLEN - 1, "%s/%s", fma_global_path, opts[i]);
            gl (gpath, FALSE, FALSE, FALSE, strlen (fma_global_path));

            ct++;

        }

    }
    else {

        while ((ent = readdir (dir)) != NULL) {

            strncpy (gpath, ent->d_name, STRLEN - 1);
            
            if (gpath[0] == '^' && gpath[1] != '$') {

                snprintf (gpath, STRLEN - 1, "%s/%s", fma_global_path, ent->d_name);

                gl (gpath, FALSE, FALSE, FALSE, strlen (fma_global_path));
                
                ct++;

            }

        }

    }

    printf ("\n\n");
    printf ("    - %d globals examined\n\n", ct);

    return 0;

}

int fma_globals_remove (int optc, char **opts)
{
    
    char gpath[STRLEN];
    int i;
    int ct = 0;
    int er = 0;
    int tot = 0;

    printf ("\nFreeM Global Removal\n");
    printf ("--------------------\n\n");

    printf ("Namespace:     %s\n", fma_namespace);
    printf ("Global Path:   %s\n\n", fma_global_path);

    for (i = 1; i < optc; i++) {
        printf ("%-10s\t", opts[i]);

        snprintf (gpath, STRLEN - 1, "%s/%s", fma_global_path, opts[i]);

        if (unlink (gpath) == -1) {
            printf ("[FAIL]\n");
            er++;
        }
        else {
            printf ("[OK]\n");
            ct++;
        }

        tot++;
    
    }

    printf ("\nRemoved %d globals [%d errors/%d attempted]\n\n", ct, er, tot);

    return 0;

}



/* PRIVATE FUNCTIONS */

/***
 * gl(): list global
 *
 *  global:  name of global
 *  kf:      key flag
 *  df:      data flag
 *  nr:      naked reference flag
 */
void gl (char *global, short kf, short df, short nr, int chop)
{
    short   filedes;            /* file descriptor */
    char    block[BLOCKLEN];    /* block to be read */
    char    key[512];
    char    prevkey[512];       /* previous key */
    char    data[1024];         /* if data has CTRLs it may become */
                                /* so long                         */
    unsigned long blknbr;
    short   offset;
    short   length;
    short   koffs;
    short   dkf = TRUE;

    short   CtrlFlag;
    short   n, k1;
    register int i, j, k, ch;

    if ((filedes = open (global, 0)) == -1) {
        printf ("%s: cannot open\012\015", global);

        return;
    }
    
    if (kf == FALSE && df == FALSE) {
        kf = TRUE;
        df = TRUE;
    } 
    else {
        dkf = FALSE;
    }

    blknbr = ROOT;
    prevkey[0] = 0;
    prevkey[1] = 0;
    prevkey[2] = 0;

    for (;;) {

        lseek (filedes, blknbr * BLOCKLEN, 0);
        
        if (read (filedes, block, BLOCKLEN) == 0) {
            fprintf (stderr, "\015*** something wrong ***\033[K\n\r");
            exit (0);
        }
        
        if (block[BTYP] == DATA) goto first;
        
        i = UNSIGN (block[0]) + 2;
        blknbr = UNSIGN (block[i]) * 65536 + UNSIGN (block[i + 1]) * 256 + UNSIGN (block[i + 2]);

    }

again:

    if (blknbr == 0) {
        close (filedes);
        return;
    }

    lseek (filedes, blknbr * 1024L, 0);
    read (filedes, block, BLOCKLEN);

first:

    offset = UNSIGN (block[OFFS]) * 256 + UNSIGN (block[OFFS + 1]);
    blknbr = UNSIGN (block[RLPTR]) * 65536 + UNSIGN (block[RLPTR + 1]) * 256 + UNSIGN (block[RLPTR + 2]);

    i = 0;
    while (i < offset) {

        length = UNSIGN (block[i++]);
        k = koffs = UNSIGN (block[i++]);
        
        if ((i + length) > offset) break;

        for (j = 0; j < length; j++) key[k++] = block[i++];
        
        key[k] = g_EOL;
        {

            short   ch0, i, j, k;

            j = 0;
            i = 0;
            data[j++] = '(';
            k = 1;
            
            while ((ch = UNSIGN (key[i++])) != g_EOL) {
            
                if (k) {
                    k = 0;
                    if (ch > SP) data[j++] = '"';
                }

                ch0 = (ch >= SP ? (ch >> 1) :       /* 'string' chars */
                    (ch < 20 ? (ch >> 1) + '0' :    /* 0...9          */
                    (ch >> 1) + SP));               /* '.' or '-'     */
                
                if (ch0 == DEL) {
                    if ((ch0 = (UNSIGN (key[i++]) >> 1)) == DEL) {
                        ch0 = (UNSIGN (key[i++]) >> 1) + DEL;
                    }
                    
                    ch0 += DEL;
                    data[j] = '<';
                    data[++j] = '0' + ch0 / 100;
                    data[++j] = '0' + (ch0 % 100) / 10;
                    data[++j] = '0' + ch0 % 10;
                    data[++j] = '>';

                } 
                else {
                    data[j] = ch0;
                }

                if (data[j++] == '"') data[j++] = '"';

                if (ch & 01) {
                    
                    if (ch > SP) data[j++] = '"';
                    
                    data[j++] = ',';
                    k = 1;

                }

            }

            data[j--] = 0;
            data[j] = ')';
            
            if (j == 0) data[0] = 0;

            while (j >= 0) {
                if ((ch = data[--j]) < SP || ch >= DEL) break;
            }

            k1 = 0;
            if (nr) {

                if (prevkey[0]) {

                    n = ch = 0;

                    while (data[n] == prevkey[n]) {
                        
                        if (prevkey[n] == '"') ch = !ch;
                        if (!ch && k1 == 0 && (prevkey[n] == '(')) k1 = n + 1;
                        if (!ch && (prevkey[n] == ',')) k1 = n + 1;

                        n++;

                    }

                    while (prevkey[n]) {

                        if (prevkey[n] == '"') ch = !ch;
                        
                        if (!ch && (prevkey[n] == ',')) {
                            k1 = 0;
                            break;
                        }

                        n++;

                    }

                }

                strcpy (prevkey, data);

                if (k1 > 1) {
                    strcpy (&data[1], &data[k1]);
                }

            }

            if (j < 0) {

                if (kf) {
                    
                    if (k1) {
                        printf ("%c%s", '^', data);
                    } 
                    else {
                        printf ("%s%s", global + chop + 1, data);
                    }

                }

                if (dkf && !nr) {
                    printf ("=");
                }
                else if (kf) {
                    printf ("\n");
                }

            } 
            else {
                fprintf (stderr, "[%d][%d] <illegal subscipt>\n", length, koffs);
            }

        }

        length = UNSIGN (block[i++]);
        
        stcpy0 (data, &block[i], (long) length);
        
        data[length] = EOL;
        
        if (numeric (data)) {
            data[length] = 0;
            i += length;
        } 
        else {

            CtrlFlag = 0;
            data[0] = '"';
            k = 1;

            while (length-- > 0) {
                
                ch = UNSIGN (block[i++]);
                
                if ((ch >= SP) && (ch < DEL)) {
                    
                    if (CtrlFlag) { /* close Bracket after CTRL */
                        
                        data[k++] = ')';
                        data[k++] = '_';
                        data[k++] = '"';

                        CtrlFlag = 0;

                    }

                    if ((data[k++] = ch) == '"') data[k++] = ch;

                } 
                else {
                    
                    if (((ch >= NUL) && (ch < SP)) || ch == DEL) {
                        
                        if (CtrlFlag) {
                            data[k++] = ',';
                        } 
                        else {

                            if (k > 1) {
                                data[k++] = '"';
                                data[k++] = '_';
                            } 
                            else {
                                k = 0;
                            }

                            data[k++] = '$';
                            data[k++] = 'C';
                            data[k++] = '(';

                            CtrlFlag = 1;

                        }
                        
                        if (ch == DEL) {
                            data[k++] = '1';
                            ch -= 100;
                        }

                        if (ch >= 10) {
                            data[k++] = ch / 10 + '0';
                            ch = ch % 10;
                        }

                        data[k++] = ch + '0';

                    } 
                    else {

                        if (CtrlFlag) { /* close Bracket after CTRL */

                            data[k++] = ')';
                            data[k++] = '_';
                            data[k++] = '"';

                            CtrlFlag = 0;

                        }

                        data[k++] = '<';

                        if (ch > 99) {
                            data[k++] = '0' + (ch / 100);
                            ch = ch % 100;
                        }

                        if (ch > 9) {
                            data[k++] = '0' + (ch / 10);
                            ch = ch % 10;
                        }

                        data[k++] = '0' + ch;
                        data[k++] = '>';

                    }
                }
            }

            if (CtrlFlag) {
                data[k++] = ')';
            }
            else {
                data[k++] = '"';
            }
            
            data[k] = 0;
        
        }

        if (df) printf ("%s\n", data);

    }

    if (i != offset) fprintf (stderr, "\nwrong offset %d vs. %d\n", offset, i);
    
    goto again;

}